* Program....: FoxMenu.prg
* Notice.....: Copyright (c) 2002 Microsoft Corp.
* Compiler...: Visual FoxPro 8.0
* Abstract...:
*	Wraps context-sensitive (right-click) menu functionality in an object.
* Changes....:

#define TESTMODE .T.

#if TESTMODE
o = CREATEOBJECT("ContextMenu")

oMenuItem = CREATEOBJECT("MenuItem")
oMenuItem.Caption = "\<New"
o.AddMenu(oMenuItem)

o.AddMenu("\<Edit", "MessageBox('hello')")
oMenuItem = o.AddMenu("\<Projects")
oMenuItem.SubMenu.AddMenu("Project 1")
oMenuItem.SubMenu.AddMenu("Project 2")

oMenuItem = o.AddMenu("\-")
oMenuItem = o.AddMenu("Test")

o.Show()
#endif


DEFINE CLASS ContextMenu AS Custom
	MenuItems = .NULL.
	MenuBarCount = 0

	PopupNames   = .NULL.
	
	PROCEDURE Init()
		THIS.MenuItems  = CREATEOBJECT("Collection")
		THIS.PopupNames = CREATEOBJECT("Collection")
	ENDPROC

	FUNCTION MenuBarCount_Access()
		RETURN THIS.MenuItems.Count
	ENDFUNC
	
	FUNCTION AddMenu(cCaption, cActionCode, cPicture, lChecked, lEnabled, lBold)
		LOCAL oMenuItem
		
		IF PCOUNT() < 5
			lEnabled = .T.
		ENDIF
		
		* we could pass a menu object rather than a caption
		* (this is our technique for overloading a function!)
		IF VARTYPE(cCaption) == 'O'
			oMenuItem = cCaption
		ELSE
			* don't add 2 menu separators in a row
			IF m.cCaption == "\-" AND THIS.MenuItems.Count > 0 AND THIS.MenuItems.Item(THIS.MenuItems.Count).Caption == "\-"
				RETURN .NULL.
			ENDIF
		
			oMenuItem = CREATEOBJECT("MenuItem")
			WITH oMenuItem
				oMenuItem.Caption = cCaption
				IF VARTYPE(cPicture) == 'C'
					.Picture = cPicture
				ENDIF
				IF VARTYPE(lChecked) == 'L'
					.Checked = lChecked
				ENDIF
				IF VARTYPE(cActionCode) == 'C'
					.ActionCode = cActionCode
				ENDIF
				IF VARTYPE(lEnabled) == 'L'
					.IsEnabled = lEnabled
				ENDIF
				IF VARTYPE(lBold) == 'L'
					.Bold = lBold
				ENDIF
			ENDWITH
		ENDIF
		THIS.MenuItems.Add(oMenuItem)
		
		RETURN oMenuItem
	ENDFUNC
	

	PROCEDURE Show(nRow, nCol)
		LOCAL nBar
		LOCAL nCol
		LOCAL nRow
		LOCAL oMenuItem
		LOCAL cActionCode
		LOCAL cPopup

		IF VARTYPE(nRow) <> 'N'
			nRow = MROW()
		ENDIF
		IF VARTYPE(nCol) <> 'N'
			nCol = MCOL()
		ENDIF

		THIS.PopupNames.Remove(-1)
		THIS.BuildMenu("shortcut", nRow, nCol)

		ACTIVATE POPUP shortcut

		FOR EACH cPopup IN THIS.PopupNames
			RELEASE POPUP &cPopup
		ENDFOR
	ENDPROC

	* Render the menu
	FUNCTION BuildMenu(cMenuName, nRow, nCol)
		LOCAL nBar
		LOCAL oMenuItem
		LOCAL cActionCode
		LOCAL cSubMenu
		LOCAL cSkipFor
		LOCAL cStyle
		
		IF VARTYPE(cMenuName) <> 'C'
			cMenuName = SYS(2015)
		ENDIF
	
		IF PCOUNT() < 3
			DEFINE POPUP (cMenuName) SHORTCUT RELATIVE
		ELSE
			DEFINE POPUP (cMenuName) SHORTCUT FROM nRow, nCol
		ENDIF
		THIS.PopupNames.Add(cMenuName)

		m.nBar = 0
		FOR EACH oMenuItem IN THIS.MenuItems
			cActionCode = oMenuItem.ActionCode

			IF oMenuItem.IsEnabled
				cSkipFor = ''
			ELSE
				cSkipFor = "SKIP FOR .T."
			ENDIF

			IF oMenuItem.Bold
				cStyle = [STYLE "B"]
			ELSE
				cStyle = ''
			ENDIF
			
			m.nBar = m.nBar + 1
			DEFINE BAR (m.nBar) OF (cMenuName) PROMPT (oMenuItem.Caption) PICTURE (oMenuItem.Picture) &cStyle &cSkipFor
			
			IF VARTYPE(cActionCode) == 'C' AND !EMPTY(cActionCode)
				ON SELECTION BAR (m.nBar) OF (cMenuName) &cActionCode
			ENDIF
			
			IF oMenuItem.Checked
				SET MARK OF BAR (m.nBar) OF (cMenuName) TO .T.
			ENDIF
			
			IF oMenuItem.SubMenu.MenuItems.Count > 0
				cSubMenu = SYS(2015)
				
				ON BAR (m.nBar) OF (cMenuName) ACTIVATE POPUP &cSubMenu

				oMenuItem.SubMenu.BuildMenu(cSubMenu)
			ENDIF
		ENDFOR
	ENDFUNC
ENDDEFINE


DEFINE CLASS MenuItem AS Custom
	Name        = "MenuItem"
	Caption     = ''
	Picture     = ''
	Checked     = .F.
	ActionCode  = ''
	IsEnabled   = .T.
	Bold        = .F.

	SubMenu = .NULL.

	PROCEDURE Init(cCaption, cActionCode, cPicture, lChecked, lEnabled)
		THIS.SubMenu = CREATEOBJECT("ContextMenu")

		IF VARTYPE(cCaption) == 'C'
			THIS.Caption = cCaption
		ENDIF
		IF VARTYPE(cPicture) == 'C'
			THIS.Picture = cPicture
		ENDIF
		IF VARTYPE(lChecked) == 'L'
			THIS.Checked = lChecked
		ENDIF
		IF VARTYPE(cActionCode) == 'C'
			THIS.ActionCode = cActionCode
		ENDIF
		IF VARTYPE(lEnabled) == 'L' AND PCOUNT() >= 5
			THIS.IsEnabled = lEnabled
		ENDIF
	ENDPROC

ENDDEFINE
